# elements.py
# -*- coding: utf-8 -*-
#
# Copyright © Stephen Day
#
# This module is part of Creoleparser and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php
#

import re
import urlparse
import urllib
import keyword
import sys

import genshi.builder as bldr
from genshi.core import Stream, Markup

from core import (escape_char, esc_neg_look, fragmentize, ImplicitList, P3Template) 

BLOCK_ONLY_TAGS = ['h1','h2','h3','h4','h5','h6',
              'ul','ol','dl',
              'pre','hr','blockquote','address',
              'p','div','form','fieldset','table',
              'noscript']

BLOCK_TAGS = BLOCK_ONLY_TAGS + ['ins','del','script']


MACRO_NAME = r'(?P<name>[a-zA-Z]+([-.]?[a-zA-Z0-9]+)*)'
"""allows any number of non-repeating hyphens or periods.
Underscore is not included because hyphen is"""


# use Genshi's HTMLSanitizer if possible (i.e., not on Google App Engine)
try:
    from genshi.filters import HTMLSanitizer
except:
    SAFE_SCHEMES = frozenset(['file', 'ftp', 'http', 'https', 'mailto', None])
    class HTMLSanitizer(object):
        def is_safe_uri(self,uri):
            if ':' not in uri:
                return True # This is a relative URI
            chars = [char for char in uri.split(':', 1)[0] if char.isalnum()]
            return ''.join(chars).lower() in SAFE_SCHEMES

sanitizer = HTMLSanitizer()

__docformat__ = 'restructuredtext en'

class WikiElement(object):
    
    """Baseclass for all wiki elements."""
    
    append_newline = False
    """Determines if newlines are appended to Element(s) during processing.
    Should only affect readability of source xml.
    """
    
    def __init__(self, tag, token, child_elements=None):
        """Constructor for WikiElement objects.

        Subclasses may have other keyword arguments.   

        :parameters:
          tag
            The xhtml tag associated with the element.
          token
            The character string (or strings) that identifies the element
            in wiki markup.
          child_elements
            A list of wiki_elements that will be searched for in the body of the
            element.  The order of these elements matters, because if an element is
            found before the element that encloses it, the enclosing element will
            never be found.  In cases where this imposes limits (e.g, ``strong`` and
            ``em`` should be allowed to nest each other), place the conflicting
            elements in a sublist. The parser will then find which comes first.
        """
        self.tag = tag
        self.token = token
        if child_elements is None:
            child_elements = []
        self.child_elements = child_elements


    def _build(self,mo,element_store, environ):
        """Returns a genshi Element that has ``self.tag`` as the
        outermost tag.

        This methods if called exclusively by ``_process``

        :parameters:
          mo
            match object, usually the one returned by
            self.regexp.search(s) 
        """
        return bldr.tag.__getattr__(self.tag)(fragmentize(mo.group(1),
                                                          self.child_elements,
                                                          element_store, environ))

    def re_string(self):
        """The regular expression pattern that is compiled into ``self.regexp``.

        The regular expression must consume the entire wiki element,
        including the tokens. For block elements, the newline on the last
        line must be consumed also. group(1) should normally be the
        entire string inside the tokens. If not, a custom ``_build``
        method will be needed.
        """
        pass

        
    def __repr__(self):
        return "<"+self.__class__.__name__ + " " + str(self.tag)+">"



    def _process(self, mos, text, wiki_elements,element_store, environ):
        """Returns genshi Fragments (Elements and text)

        This is mainly for block level markup. See InlineElement
        for the other method.
        """
        frags = []
        end = 0
        for mo in mos:
            if end != mo.start():
            # call again for leading text and extend the result list 
                frags.extend(fragmentize(text[end:mo.start()],wiki_elements[1:],
                                         element_store, environ))
            # append the found wiki element to the result list
            built = self._build(mo,element_store, environ)
            if built is not None:
                frags.append(built)
            # make the source output easier to read
            if self.append_newline:
                frags.append('\n')
            end = mo.end()
        # call again for trailing text and extend the result list
        if end < len(text):
            if not isinstance(wiki_elements[0],(list,tuple)):
                wiki_elements = wiki_elements[1:]
            frags.extend(fragmentize(text[end:],wiki_elements,
                                         element_store, environ))

        return frags


class BlockElement(WikiElement):

    """Block elements inherit form this class

    Wiki elements wanting ``append_newline = True`` should use this
    as the base also.

    """

    append_newline = True
    

class InlineElement(WikiElement):

    r"""For finding generic inline elements

    >>> em = InlineElement('em','//')
    >>> mo1 = em.regexp.search('a //word// in a line')
    >>> mo2 = em.regexp.search('a //word in a line\n or two\n')
    >>> mo1.group(0),mo1.group(1)
    ('//word//', 'word')
    >>> mo2.group(0),mo2.group(1)
    ('//word in a line\n or two', 'word in a line\n or two')

    Use a list for the ``token`` argument to have different start
    and end strings. These must be closed.

    >>> foo = InlineElement('foo',['<<','>>'])
    >>> mo = foo.regexp.search('blaa <<here it is >>\n')
    >>> mo.group(1)
    'here it is '
        
    """

    def __init__(self, tag='', token=''):
        super(InlineElement,self).__init__(tag,token)
        self.regexp = re.compile(self.re_string(),re.DOTALL)

    def re_string(self):
        if isinstance(self.token,str):
            content = '(.+?)'
            end = '(' + esc_neg_look + re.escape(self.token) + r'|$)'
            return esc_neg_look + re.escape(self.token) + content + end
        else:
            content = '(.+?)'
            return esc_neg_look + re.escape(self.token[0]) + content + esc_neg_look + re.escape(self.token[1])

    def _process(self, mos, text, wiki_elements, element_store, environ):
        """Returns genshi Fragments (Elements and text)"""
        parts = []
        end = 0
        for mo in mos:
            processed = self._build(mo,element_store, environ)
            store_id = str(id(processed)) 
            element_store[store_id] = processed
            parts.append(''.join([text[end:mo.start()],'<<<',store_id,'>>>']))
            end = mo.end()
        # call again for trailing text and extend the result list
        if end < len(text):
            parts.append(text[end:])
        new_text = ''.join(parts)
        if not isinstance(wiki_elements[0],(list,tuple)):
            wiki_elements = wiki_elements[1:]
        frags = fragmentize(new_text,wiki_elements,element_store, environ)
        return frags


class SimpleElement(InlineElement):

    r"""For finding generic inline elements like ``strong`` and ``em``.

    >>> em = SimpleElement({'//':'em'})
    >>> mo1 = em.regexp.search('a //word// in a line')
    >>> mo2 = em.regexp.search('a //word in a line\n or two\n')
    >>> mo1.group(0),mo1.group(2)
    ('//word//', 'word')
    >>> mo2.group(0),mo2.group(2)
    ('//word in a line\n or two', 'word in a line\n or two')
       
    """

    def __init__(self, token_dict={}):
        self.token_dict = token_dict
        self.tokens = token_dict.keys()
        super(SimpleElement,self).__init__('','')
        self.regexp = re.compile(self.re_string(),re.DOTALL)

    def re_string(self):
        if isinstance(self.token,basestring):
            tokens = '(' + '|'.join([re.escape(token) for token in self.tokens]) + ')'
            content = '(.+?)'
            end = '(' + esc_neg_look + r'\1|$)'
            return esc_neg_look + tokens + content + end

    def _build(self,mo,element_store, environ):
        return bldr.tag.__getattr__(self.token_dict[mo.group(1)])(fragmentize(mo.group(2),
                                                          self.child_elements,
                                                          element_store, environ))


class LinkElement(InlineElement):

    """Superclass for AnchorLinks and ImageLinks. Parses internal, external,
    and interwiki links.
    
    """
    
    def __init__(self,tag, token, delimiter,
                 interwiki_delimiter,base_urls,links_funcs,default_space_char,space_chars,
                 base_url,space_char,class_func,path_func):
        super(LinkElement,self).__init__(tag,token)
        self.regexp = re.compile(self.re_string(),re.DOTALL)
        self.delimiter = delimiter
        self.interwiki_delimiter = interwiki_delimiter
        self.base_urls = base_urls
        self.links_funcs = links_funcs
        self.default_space_char = default_space_char
        self.space_chars = space_chars
        self.base_url = base_url
        self.space_char = space_char
        self.class_func = class_func
        self.path_func = path_func
        self.content_regexp = re.compile(self.content_re_string(),re.DOTALL)
##        self.arg_regexp = re.compile(self.arg_re_string(),re.DOTALL)
        self.interwikilink_regexp = re.compile(self.interwikilink_re_string())
        self.urllink_regexp = re.compile(self.urllink_re_string(), re.DOTALL)
        self.wikilink_regexp = re.compile(self.wikilink_re_string())

##    def arg_re_string(self):
##        key = r'((?P<key>\w+)\s*\=)?'
##        value = r'(?P<value>.*?)'
##        return r'\s*' + key + r'\s*' + value + r'\s*(?P<delimiter>' + \
##               re.escape(self.delimiter) + r'|$)(?P<tail>.*)'

    def content_re_string(self):
        return r'(?P<body>.*?)(' + re.escape(self.delimiter) + '(?P<arg_string>.*?))?$'

    def interwikilink_re_string(self):
        all_wikis = set(self.links_funcs.keys() + self.base_urls.keys())
        wiki_id = '(?P<wiki_id>' + '|'.join(all_wikis) + ')'
        optional_spaces = ' *'
        page_name = r'(?P<page_name>\S+?( \S+?)*)' #allows any number of single spaces
        return '^' + optional_spaces + wiki_id + \
               re.escape(self.interwiki_delimiter) + ' *' + page_name + \
               optional_spaces + '$'#+ alias

    def urllink_re_string(self):
        protocol = r'^\s*((\w+?:|/)' 
        rest_of_url = r'[\S\n]*?)\s*$'
        return protocol + rest_of_url #+ alias

    def wikilink_re_string(self):
        optional_spaces = ' *'
        page_name = r'(?P<page_name>\S+?( \S+?)*?)' #allows any number of single spaces
        return '^' + optional_spaces + page_name + optional_spaces + '$'#+ \

##    def parse_args(self, arg_string):
##        args = []
##        delimiter = True
##        while delimiter:
##            mo = self.arg_regexp.match(arg_string)
##            key, value, delimiter, tail = mo.group('key'),mo.group('value'),mo.group('delimiter'), mo.group('tail')
##            if key:
##                args.append((key, value))
##            else:
##                args.append(value)
##            arg_string = tail
##        positional_args = []
##        kw_args = {}
##        for arg in args:
##           if isinstance(arg,tuple):
##             k, v  = arg
##             k = str(k).lower()
##             if k in keyword.kwlist:
##                 k = k + '_'
##             if k in kw_args:
##                if isinstance(v,list):
##                   try:
##                      kw_args[k].extend(v)
##                   except AttributeError:
##                      v.insert(0,kw_args[k])
##                      kw_args[k] = v
##                elif isinstance(kw_args[k],list):
##                   kw_args[k].append(v)
##                else:
##                   kw_args[k] = [kw_args[k], v]
##                kw_args[k] = ImplicitList(kw_args[k])
##             else:
##                kw_args[k] = v
##             if isinstance(kw_args[k],ImplicitList):
##                 kw_args[k] = ','.join(kw_args[k])
##           else:
##             positional_args.append(arg)
##
##        return (positional_args, kw_args)


    def page_name(self,mo):
        if 'wiki_id' in mo.groupdict():
            space_char = self.space_chars.get(mo.group('wiki_id'),self.default_space_char)
        else:
            space_char = self.space_char
        return mo.group('page_name').replace(' ',space_char)

    def _build(self,mo,element_store, environ):
        content_mo = self.content_regexp.match(mo.group(1))
        body = content_mo.group('body')
        arg_string = content_mo.group('arg_string')
        the_class = None
        page_name = None
        if self.interwikilink_regexp.match(body):
            interwikilink_mo = self.interwikilink_regexp.match(body)
            link_type = 'interwiki'
            base_url = self.base_urls.get(interwikilink_mo.group('wiki_id'))
            link_func = self.links_funcs.get(interwikilink_mo.group('wiki_id'))
            page_name = self.page_name(interwikilink_mo)
            if link_func:
                url = link_func(page_name)
            else:
                url = urllib.quote(page_name.encode('utf-8'))
            if base_url:
                url = urlparse.urljoin(base_url, url)
        elif self.urllink_regexp.match(body):
            urllink_mo = self.urllink_regexp.match(body)
            link_type = 'url'
            if sanitizer.is_safe_uri(urllink_mo.group(1)):
                url = urllink_mo.group(1)
            else:
                url = None
        elif self.wikilink_regexp.match(body):
            wikilink_mo = self.wikilink_regexp.match(body)
            link_type = 'wiki'
            page_name = self.page_name(wikilink_mo)
            if self.path_func:
                the_path = self.path_func(self.tag, page_name, environ)
            else:
                the_path = urllib.quote(page_name.encode('utf-8'))
            url = urlparse.urljoin(self.base_url, the_path)
        else:
            url = None

        if not url:
            return mo.group(0)
        else:
            if arg_string is not None:
                args, kw_args = [arg_string.strip()], {} #self.parse_args(arg_string)
            else:
                args, kw_args = [], {}

            try:
                if self.class_func:
                    the_class = self.class_func(link_type, url, body, page_name)
                return self.emit(element_store, environ,link_type,body,url,the_class, *args, **kw_args)
            except TypeError:
                return mo.group(0)
                


class AnchorElement(LinkElement):

    """Finds and builds internal, external, and interwiki links.

    >>> link = AnchorElement('a',('[[',']]'),'|',
    ... interwiki_delimiter=':', 
    ... base_urls=dict(somewiki='http://somewiki.org/',
    ...                bigwiki='http://bigwiki.net/'),
    ... links_funcs={},default_space_char='-',
    ... space_chars={'bigwiki':' '},base_url='http://somewiki.org/',
    ... space_char='_',class_func=None,path_func=None)
    
    >>> mo = link.regexp.search("[[http://www.google.com| here]]")
    >>> link._build(mo,{},None).generate().render()
    '<a href="http://www.google.com">here</a>'

    >>> mo = link.regexp.search(" [[somewiki:Home Page|steve]] ")
    >>> link._build(mo,{},None).generate().render()
    '<a href="http://somewiki.org/Home-Page">steve</a>'

    >>> mo = link.regexp.search(" [[bigwiki:Home Page]] ")
    >>> link._build(mo,{},None).generate().render()
    '<a href="http://bigwiki.net/Home%20Page">bigwiki:Home Page</a>'

    >>> mo = link.regexp.search(" [[Home Page |Home]]")
    >>> link._build(mo,{},None).generate().render()
    '<a href="http://somewiki.org/Home_Page">Home</a>'
    
    """
    

    def __init__(self, *args, **kw_args):
        super(AnchorElement,self).__init__(*args, **kw_args)    

    def emit(self,element_store, environ,link_type,body,url,the_class, alias=None):
        if alias:
            alias = fragmentize(alias,self.child_elements,element_store, environ)
        else:
            alias = body.strip()
        return bldr.tag.__getattr__(self.tag)(alias,
                                              href=url,
                                              class_=the_class)

        
class ImageElement(LinkElement):

    def __init__(self, *args, **kw_args):
        super(ImageElement,self).__init__(*args, **kw_args)    

    def emit(self,element_store, environ,link_type,body,url,the_class, alt=None):
        if alt is None:
            if link_type == 'url':
                alt = urlparse.urlsplit(url).path.split('/')[-1]
            else:
                alt = body.strip()
        return bldr.tag.__getattr__(self.tag)(src=url ,alt=alt, title=alt,
                                              #class_=the_class
                                              )           
        

class Link(InlineElement):

    """Finds and builds links."""
    
    def __init__(self,tag, token):
        super(Link,self).__init__(tag,token)
        self.regexp = re.compile(self.re_string(),re.DOTALL)

    def _build(self,mo,element_store, environ):
        
        for tag in self.child_elements:
            m = tag.regexp.search(mo.group(1))
            if m:
                link = tag._build(m,element_store, environ)
                if link:
                    break
        else:
            link = None

        if link:
            return bldr.tag(link)
        else:
            return mo.group(0)


class Macro(WikiElement):
    r"""Finds and processes inline macro elements."""

    def __init__(self, tag, token, func):
        super(Macro,self).__init__(tag,token , [])
        self.func = func
        self.regexp = re.compile(self.re_string())


    def _process(self, mos, text, wiki_elements,element_store, environ):
        """Returns genshi Fragments (Elements and text)"""
        assert len(mos) == 1
        mo = mos[0]
        processed = self._build(mo,element_store, environ)
        if isinstance(processed, list):
            tail = processed[1]
            processed = processed[0]
        else:
            tail = ''
        if isinstance(processed, basestring) and not isinstance(processed,Markup):
            text = ''.join([text[:mo.start()],processed,tail,
                        text[mo.end():]])
        else:
            store_id = str(id(processed))
            element_store[store_id] = processed
            text = ''.join([text[:mo.start()],'<<<',store_id,'>>>',tail,
                        text[mo.end():]])
        frags = fragmentize(text,wiki_elements,element_store, environ)
        return frags


    def re_string(self):
        content = '(.*?)'
        return esc_neg_look + re.escape(self.token[0]) + r'(' + MACRO_NAME + \
               content + ')' + esc_neg_look + re.escape(self.token[1])

    trailing_slash = re.compile(r'(?<=[ "\'\]])/$')
    def _build(self,mo,element_store, environ):
        arg_string = re.sub(self.trailing_slash,'',mo.group(4))
        if self.func:
            value = self.func(mo.group('name'),arg_string,None,False,environ)
        else:
            value = None
        if value is None:
            return bldr.tag.code(self.token[0],bldr.tag.span(mo.group('name'),class_="macro_name"),
                           bldr.tag.span(arg_string,class_="macro_arg_string"),
                           self.token[1],class_="unknown_macro")
        elif isinstance(value, (basestring,bldr.Fragment,bldr.Element, Stream)):
            return value
        else:
            raise Exception("macros can only return strings and genshi objects") 
        

class BodiedMacro(Macro):
    """Finds and processes macros with bodies.

    Does not span across top level block markup
    (see BodiedBlockMacro's for that)."""

    def __init__(self, tag, token, func):
        super(BodiedMacro,self).__init__(tag,token , func)
        self.func = func
        self.regexp = re.compile(self.re_string(),re.DOTALL)

    def re_string(self):
        content = r'(?P<arg_string>[ \S]*?)'
        body = '(?P<body>.+)'
        return esc_neg_look + re.escape(self.token[0]) + MACRO_NAME + \
               content + '(?<!/)' + re.escape(self.token[1]) + \
               body + esc_neg_look + re.escape(self.token[0]) + \
               r'/(?P=name)' + '(?<!/)' + re.escape(self.token[1])

    def _build(self,mo,element_store, environ):
        start = ''.join([esc_neg_look, re.escape(self.token[0]), re.escape(mo.group('name')),
                         r'(?P<arg_string>[ \S]*?)', re.escape(self.token[1])])
        end = ''.join([esc_neg_look, re.escape(self.token[0]), '/', re.escape(mo.group('name')),
                       re.escape(self.token[1])])
        count = 0
        for mo2 in re.finditer(start + '|' + end, mo.group('body')):
            if re.match(end,mo2.group(0)):
                count = count + 1
            else:
                count = count - 1
            if count > 0:
                body = mo.group('body')[:mo2.start()]
                tail = ''.join([mo.group('body')[mo2.end():], self.token[0],
                                '/', mo.group('name'), self.token[1]])
                break
        else:
            body = mo.group('body')
            tail = ''
                
                
        
        if self.func:
            value = self.func(mo.group('name'),mo.group('arg_string'),body,False,environ)
        else:
            value = None
        if value is None:
            content_out = [self.token[0],bldr.tag.span(mo.group('name'),class_="macro_name"),
                           bldr.tag.span(mo.group('arg_string'),class_="macro_arg_string"),
                           self.token[1],bldr.tag.span(mo.group('body'),class_="macro_body"),
                           self.token[0] + '/' + mo.group('name') + self.token[1]]
            return [bldr.tag.code(content_out,class_="unknown_macro", style="white-space:pre-wrap"),tail]
            
        elif isinstance(value, (basestring,bldr.Fragment, Stream)):
            return [value,tail]
        else:
            raise Exception("macros can only return strings and genshi objects")

       

class BodiedBlockMacro(WikiElement):
    """Finds and processes block macros with bodies.

    The opening and closing tokens must be are each on a line alone without
    leading spaces. These macros can enclose other block level markup
    including pre blocks and other BodiedBlockMacro's."""


    def __init__(self, tag, token, func):
        super(BodiedBlockMacro,self).__init__(tag,token , func)
        self.func = func
        self.regexp = re.compile(self.re_string(),re.DOTALL+re.MULTILINE)

    def re_string(self):
        arg_string = r'(?P<arg_string>(?![^\n]*>>[^\n]*>>)[ \S]*?)'
        start = '^' + re.escape(self.token[0])
        body = r'(?P<body>.*\n)'
        end = re.escape(self.token[0]) + \
               r'/(?P=name)' + '(?<!/)' + re.escape(self.token[1]) + r'\s*?$'
        
        return start + '(' + MACRO_NAME + arg_string + ')' + '(?<!/)' + \
               re.escape(self.token[1]) + r'\s*?\n' + body + end

    def _process(self, mos, text, wiki_elements,element_store, environ):
        """Returns genshi Fragments (Elements and text)

        This is mainly for block level markup. See InlineElement
        for the other method.
        """
        assert len(mos) == 1
        mo = mos[0]
        processed = self._build(mo,element_store, environ)
        if isinstance(processed, list):
            tail = processed[1]
            processed = processed[0]
        else:
            tail = ''
        if isinstance(processed, basestring) and not isinstance(processed,Markup):
            #print '_process', repr(processed)
            text = ''.join([text[:mo.start()],processed,tail,
                        text[mo.end():]])
            frags = fragmentize(text,wiki_elements,element_store, environ)
        else:
        
            frags = []
            # call again for leading text and extend the result list 
            if mo.start():
                frags.extend(fragmentize(text[:mo.start()],wiki_elements[1:],
                                         element_store, environ))
            # append the found wiki element to the result list
            frags.append(processed)
            # make the source output easier to read
            if self.append_newline:
                frags.append('\n')
            # call again for trailing text and extend the result list
            if tail or mo.end() < len(text):
                frags.extend(fragmentize(tail + text[mo.end():],wiki_elements,
                                         element_store, environ))
        return frags

    def _build(self,mo,element_store, environ):
        start = ''.join(['^', re.escape(self.token[0]), re.escape(mo.group('name')),
                         r'(?P<arg_string>(?![^\n]*>>[^\n]*>>)[ \S]*?)', re.escape(self.token[1]),r'\s*?\n'])
        end = ''.join(['^', re.escape(self.token[0]), '/', re.escape(mo.group('name')),
                       re.escape(self.token[1]),r'\s*?$'])
        count = 0
        for mo2 in re.finditer(start + '|' + end, mo.group('body'),re.MULTILINE):
            if re.match(end,mo2.group(0)):
                count = count + 1
            else:
                count = count - 1
            if count > 0:
                body = mo.group('body')[:mo2.start()]
                tail = ''.join([mo.group('body')[mo2.end():], self.token[0],
                                '/', mo.group('name'), self.token[1],'\n'])
                break
        else:
            body = mo.group('body')
            tail = ''

        if self.func:
            value = self.func(mo.group('name'),mo.group('arg_string'),body,True,environ)
        else:
            value = None
        if value is None:
            return [bldr.tag.pre(self.token[0],bldr.tag.span(mo.group('name'),class_="macro_name"),
                           bldr.tag.span(mo.group('arg_string'),class_="macro_arg_string"),
                           self.token[1],'\n',bldr.tag.span(mo.group('body'),class_="macro_body"),
                           self.token[0] + '/' + mo.group('name') + self.token[1],
                           class_="unknown_macro"), tail]
        elif (isinstance(value, (Stream, basestring)) or
             (isinstance(value,bldr.Element) and value.tag in BLOCK_TAGS)):
            return [value, tail]
        # Add a p tag if the value is a Fragment or Element that needs one
        elif isinstance(value, bldr.Fragment):
            return [bldr.tag.p(value), tail]
        else:
            raise Exception("macros can only return strings and genshi objects")
        
    
class RawLink(InlineElement):
    
    """Used to find raw urls in wiki text and build xml from them.

    >>> raw_link = RawLink(tag='a')
    >>> mo = raw_link.regexp.search(" a http://www.google.com url ")
    >>> raw_link.href(mo)
    'http://www.google.com'
    >>> raw_link._build(mo,{},None).generate().render()
    '<a href="http://www.google.com">http://www.google.com</a>'
    
    """
    linking_protocols = ['http','https']
    
    def __init__(self, tag):
        super(RawLink,self).__init__(tag=tag, token=None)
        self.regexp = re.compile(self.re_string())

    def re_string(self):
        escape = '(' + re.escape(escape_char) + ')?'
        #protocol = '((https?|ftp)://'
        protocol = '((https?)://'
        rest_of_url = r'\S+?)'
        #allow one punctuation character or '**' or '//'. Don't include a placeholder.
        look_ahead = r'(?=([>)}\]]?[,.?!:;"\']?(([^a-zA-Z0-9])\6)?(\s|$))|<<<)'
        return escape + protocol + rest_of_url + look_ahead

    def _build(self,mo,element_store, environ):
        if (not mo.group(1)) and (mo.group(3) in self.linking_protocols):
            return bldr.tag.__getattr__(self.tag)(self.alias(mo,element_store),
                                              href=self.href(mo))
        else:
            return self.href(mo)
        
    def href(self,mo):
        """Returns the string for the href attribute of the Element."""
        if sanitizer.is_safe_uri(mo.group(2)):
            return mo.group(2)
        else:
            return "unsafe_uri_detected"

    def alias(self,mo,element_store):
        """Returns the string for the content of the Element."""
        return self.href(mo)


class URLLink(WikiElement):
    
    """Used to find url type links inside a link.

    The scope of these is within link markup only (i.e., [[url]]

    >>> url_link = URLLink('a','|')
    >>> mo = url_link.regexp.search(" http://www.google.com| here ")
    >>> url_link.href(mo)
    'http://www.google.com'
    >>> url_link._build(mo,{},None).generate().render()
    '<a href="http://www.google.com">here</a>'
    
    """

    def __init__(self, tag,delimiter):
        super(URLLink,self).__init__(tag, '')
        self.delimiter = delimiter
        self.regexp = re.compile(self.re_string(),re.DOTALL)

    def re_string(self):
        protocol = r'^\s*((\w+?:|/)' 
        rest_of_url = r'[\S\n]*?)\s*'
        alias = r'(' + re.escape(self.delimiter) + r' *(.*?))? *$'
        return protocol + rest_of_url + alias

    def _build(self,mo,element_store, environ):
        if not self.href(mo):
            return None
        return bldr.tag.__getattr__(self.tag)(self.alias(mo,element_store, environ),
                                              href=self.href(mo))
       
    def href(self,mo):
        """Returns the string for the href attribute of the Element."""
        if sanitizer.is_safe_uri(mo.group(1)):
            return mo.group(1)
        else:
            return None 
            

    def alias(self,mo,element_store, environ):
        """Returns the string for the content of the Element."""
        if not mo.group(4):
            return self.href(mo)
        else:
            return fragmentize(mo.group(4),self.child_elements,element_store, environ)



class InterWikiLink(WikiElement):

    """Used to match interwiki links inside a link.

    The search scope for these is only inside links. 

    >>> interwiki_link = InterWikiLink('a',
    ... delimiter1=':', delimiter2 = '|',
    ... base_urls=dict(somewiki='http://somewiki.org/',
    ...                bigwiki='http://bigwiki.net/'),
    ...                links_funcs={},default_space_char='_',
    ...                space_chars={})
    >>> mo = interwiki_link.regexp.search(" somewiki:Home Page|steve ")
    >>> interwiki_link.href(mo)
    'http://somewiki.org/Home_Page'
    >>> interwiki_link.alias(mo,{},None)
    ['steve']
    
    """

    def __init__(self, tag, delimiter1,
                 delimiter2,base_urls,links_funcs,default_space_char,space_chars):
        super(InterWikiLink,self).__init__(tag, '')
        self.delimiter1 = delimiter1
        self.delimiter2 = delimiter2
        #self.regexp = re.compile(self.re_string())
        self.base_urls = base_urls
        self.links_funcs = links_funcs
        self.default_space_char = default_space_char
        self.space_chars = space_chars
        self.regexp = re.compile(self.re_string())

    def re_string(self):
        #all_wikis = set(self.links_funcs.keys() + self.base_urls.keys())
        #wiki_id = '(' + '|'.join(all_wikis) + ')'

        wiki_id = r'(\w+)'
        optional_spaces = ' *'
        page_name = r'(\S+?( \S+?)*)' #allows any number of single spaces
        alias = r'(' + re.escape(self.delimiter2) + r' *(.*?))? *$'
        return '^' + optional_spaces + wiki_id + optional_spaces + \
               re.escape(self.delimiter1) + optional_spaces + page_name + \
               optional_spaces + alias

    def page_name(self,mo):
        space_char = self.space_chars.get(mo.group(1),self.default_space_char)
        return mo.group(2).replace(' ',space_char)

    def href(self,mo):
        linktype = mo.group(1)
        base_url = self.base_urls.get(linktype)
        link_func = self.links_funcs.get(linktype)
        if not (link_func or base_url):
            return None
        else:
            href = self.page_name(mo)
            if link_func:
                href = link_func(href)
            else:
                href = urllib.quote(href.encode('utf-8'))
            if base_url:
                href = urlparse.urljoin(base_url, href)
            return href

    def _build(self,mo,element_store, environ):
        if not self.href(mo):
            return '[[' + mo.group(0) + ']]'
        return bldr.tag.__getattr__(self.tag)(self.alias(mo,element_store, environ),
                                              href=self.href(mo))
    def alias(self,mo,element_store, environ):
        """Returns the string for the content of the Element."""
        if not mo.group(5):
            return ''.join([mo.group(1),self.delimiter1,mo.group(2)])
        else:
            return fragmentize(mo.group(5),self.child_elements,element_store, environ)



class WikiLink(WikiElement):

    """Used to match wiki links inside a link.

    The search scope for these is only inside links.

    >>> wiki_link = WikiLink('a','|',base_url='http://somewiki.org/',
    ...                      space_char='_',class_func=None, path_func=None)
    >>> mo = wiki_link.regexp.search(" Home Page |Home")
    >>> wiki_link.href(mo)
    'http://somewiki.org/Home_Page'
    >>> wiki_link.alias(mo,{},None)
    ['Home']
    
    """

    def __init__(self, tag, delimiter,
                 base_url,space_char,class_func,path_func):
        super(WikiLink,self).__init__(tag, '')
        self.delimiter = delimiter
        self.base_url = base_url
        self.space_char = space_char
        self.class_func = class_func
        self.path_func = path_func
        self.regexp = re.compile(self.re_string())

    def re_string(self):
        optional_spaces = ' *'
        page_name = r'(\S+?( \S+?)*?)' #allows any number of single spaces
        alias = r'(' + re.escape(self.delimiter) + r' *(.*?))? *$'
        return '^' + optional_spaces + page_name + optional_spaces + \
               alias

    def page_name(self,mo):
        return mo.group(1).replace(' ',self.space_char)
    
    def href(self,mo,environ):
        if self.path_func:
            the_path = self.path_func(self.tag, self.page_name(mo), environ)
        else:
            the_path = urllib.quote(self.page_name(mo).encode('utf-8'))
        return urlparse.urljoin(self.base_url, the_path)

    def _build(self,mo,element_store, environ):
        if self.class_func:
            the_class = self.class_func(self.page_name(mo))
        else:
            the_class = None
        return bldr.tag.__getattr__(self.tag)(self.alias(mo,element_store, environ),
                                              href=self.href(mo, environ),
                                              class_=the_class)
    
    def alias(self,mo,element_store, environ):
        """Returns the string for the content of the Element."""
        if not mo.group(3):
            return mo.group(1)
        else:
            return fragmentize(mo.group(4),self.child_elements,element_store, environ)


class List(BlockElement):

    """Finds list (ordered, unordered, and definition) wiki elements.

    group(1) of the match object includes all lines from the list
    including newline characters.
        
    """

    def __init__(self, tag, token,stop_tokens=None):
        self.stop_tokens = stop_tokens
        super(List,self).__init__(tag, token)
        #self.stop_tokens = stop_tokens
        self.regexp = re.compile(self.re_string(),re.DOTALL+re.MULTILINE)

    def re_string(self):
        """This re_string is for finding generic block elements like
        lists (ordered, unordered, and definition) that start with a
        single token.
        """
        leading_whitespace = r'^([ \t]*'
        only_one_token = re.escape(self.token)+ '(?!' + re.escape(self.token) + ')'
        rest_of_list = r'.*?(?:\n|\Z))'
        only_one_stop_token = '([' + re.escape(self.stop_tokens) + r'])(?!\3)'        
        look_ahead = '(?=([ \t]*' + only_one_stop_token + '|$))'
        return leading_whitespace + only_one_token + rest_of_list + \
               look_ahead



class ListItem(BlockElement):
    r"""Matches the current list item.

    Everything up to the next same-level list item is matched.

    >>> list_item = ListItem('li','#*')
    >>> mo = list_item.regexp.search("*one\n**one.1\n**one.2\n*two\n")
    >>> mo.group(3)
    'one\n**one.1\n**one.2\n'
    >>> mo.group(0)
    '*one\n**one.1\n**one.2\n'
    
    """
    
    append_newline = False

    def __init__(self, tag, list_tokens):
        """Constructor for list items.

        :parameters"
          list_tokens
            A string that includes the tokens used for lists
        """
        self.list_tokens = list_tokens
        super(ListItem,self).__init__(tag, None)
        self.regexp = re.compile(self.re_string(),re.DOTALL)

    def re_string(self):
        whitespace = r'[ \t]*'
        item_start = '(([' + self.list_tokens + r'])\2*)'
        rest_of_item = r'(.*?(?:\n|\Z))'
        start_of_same_level_item = r'\1(?!\2)'
        look_ahead = r'(?=(' + whitespace + start_of_same_level_item + '|$))'
        return whitespace + item_start + whitespace + \
               rest_of_item + look_ahead

    def _build(self,mo,element_store, environ):
        return bldr.tag.__getattr__(self.tag)(fragmentize(mo.group(3),
                                                          self.child_elements,
                                                          element_store, environ))


class NestedList(WikiElement):

    r"""Finds a list in the current list item.

    >>> nested_ul = NestedList('ul','*')
    >>> mo = nested_ul.regexp.search('one\n**one.1\n**one.2\n')
    >>> mo.group(1)
    '**one.1\n**one.2\n'
    >>> mo.group(0) == mo.group(1)
    True

    """

    def __init__(self, tag, token):
        super(NestedList,self).__init__(tag, token)
        self.regexp = re.compile(self.re_string(),re.DOTALL+re.MULTILINE)

    def re_string(self):
        look_behind = r'(?<=\n)' # have to avoid finding a list on the first line
        whitespace = r'(\s*'
        rest_of_list = '.*$)'
        return look_behind + '^' + whitespace + re.escape(self.token) + \
               rest_of_list


class DefinitionTerm(BlockElement):

    r"""Processes definition terms.

    >>> term = DefinitionTerm('dt',';',stop_token=':')
    >>> mo1,mo2 = term.regexp.finditer(";term1\n:def1\n;term2:def2\n")
    >>> mo1.group(1), mo2.group(1)
    ('term1', 'term2')
    >>> mo1.group(0), mo2.group(0)
    (';term1\n', ';term2')

    group(1) of the match object is the term line or up to the first ':'
        
    """

    def __init__(self, tag, token,stop_token):
        super(DefinitionTerm,self).__init__(tag, token)
        self.stop_token = stop_token
        self.regexp = re.compile(self.re_string(),re.DOTALL+re.MULTILINE)

    def re_string(self):
        look_ahead = r'(\n|(?=(' + esc_neg_look + re.escape(self.stop_token) + r'|$)))'
        return r'^[ \t]*' + re.escape(self.token) + r'[ \t]*(.*?' + \
               re.escape(self.stop_token) +  '?)\s*' + look_ahead 


class DefinitionDef(BlockElement):

    r"""Processes definitions.

    >>> definition = DefinitionDef('dd',':')
    >>> mo1,mo2 = definition.regexp.finditer(":def1a\ndef1b\n:def2\n")
    >>> mo1.group(1), mo2.group(1)
    ('def1a\ndef1b', 'def2')
    >>> mo1.group(0), mo2.group(0)
    (':def1a\ndef1b\n', ':def2\n')

    group(1) of the match object includes all lines from the defintion
    up to the next definition.
        
    """

    def __init__(self, tag, token):
        super(DefinitionDef,self).__init__(tag, token)
        self.regexp = re.compile(self.re_string(),re.DOTALL+re.MULTILINE)

    def re_string(self):
        look_ahead = r'(?=(^[ \t]*' + re.escape(self.token) + r')|\Z)'
        return r'^[ \t]*' + re.escape(self.token) + r'?[ \t]*(.+?)\s*' + look_ahead 

class Paragraph(BlockElement):
    """"This should be the last outer level wiki element to be searched.

    Anything that is left over will be placed in a paragraphs unless it looks
    like block content according to xhtml1 strict. Block content is defined
    here as valid children of the <body> element (see BLOCK_TAGS). Only genshi
    Element objects will be evaluated (see BLOCK_TAGS). Fragments and stings
    are treated as inline while Streams are block content.
    
    """

    def __init__(self, tag):
        super(Paragraph,self).__init__(tag,None)
        self.regexp = re.compile(self.re_string(),re.DOTALL)#+re.MULTILINE)

    def re_string(self):
        return r'^(.*?)\n?$' 

    def _build(self,mo,element_store, environ):
        content = fragmentize(mo.group(1), self.child_elements, element_store, environ)
        # Check each list item and record those that are block only
        block_only_frags = []
        for i,element in enumerate(content):
            if ((isinstance(element, bldr.Element) and
                element.tag in BLOCK_ONLY_TAGS) or
                isinstance(element,(Stream,Markup))):
                block_only_frags.append(i)
        # Build a new result list if needed
        if block_only_frags:
            new_content = []
            last_i = -1
            for i in block_only_frags:
                if content[last_i+1:i]:
                    if not (len(content[last_i+1:i])==1 and
                                                content[last_i+1] == '\n'):
                        new_content.append(bldr.tag.__getattr__(self.tag)(content[last_i+1:i]))
                    else:
                        new_content.append('\n')
                new_content.append(content[i])
                last_i = i
            if content[last_i+1:]:
                new_content.append(bldr.tag.__getattr__(self.tag)(content[last_i+1:]))
            return bldr.tag(new_content)
        else:
            return bldr.tag.__getattr__(self.tag)(content)
            

class Heading(BlockElement):

    r"""Finds heading wiki elements.

    >>> h1 = Heading(['h1','h2'],'=')
    >>> mo = h1.regexp.search('before\n = An important thing = \n after')
    >>> mo.group(2)
    'An important thing'
    >>> mo.group(0)
    ' = An important thing = \n'

    """
  
    def __init__(self, tag, token):
        super(Heading,self).__init__('',token)
        self.tags = tag
        self.regexp = re.compile(self.re_string(),re.MULTILINE)

    def re_string(self):
        whitespace = r'[ \t]*'
        tokens = '(' + re.escape(self.token) + '{1,' + str(len(self.tags)) +'})'
        content = '(.*?)'
        trailing_markup = '(' + re.escape(self.token) + r'+[ \t]*)?(\n|\Z)'
        return '^' + whitespace + tokens + \
               whitespace + content + whitespace + trailing_markup

    def _build(self,mo,element_store, environ):
        heading_tag = self.tags[len(mo.group(1))-1]
        return bldr.tag.__getattr__(heading_tag)(fragmentize(mo.group(2),
                                                          self.child_elements,
                                                          element_store, environ))


class Table(BlockElement):

    r"""Find tables.

    >>> table = Table('table','|')
    >>> mo = table.regexp.search("before\n | one | two |\n|one|two \n hi")
    >>> mo.group(1)
    ' | one | two |\n|one|two \n'
    >>> mo.group(0) == mo.group(1)
    True
    
    """

    def __init__(self, tag, token):
        super(Table,self).__init__(tag,token)
        self.regexp = re.compile(self.re_string(),re.MULTILINE)

    def re_string(self):
        whitespace = r'[ \t]*'
        rest_of_line = r'.*?(\n|\Z)'
        return '^((' + whitespace + re.escape(self.token) + \
               rest_of_line + ')+)'



class TableRow(BlockElement):

    r"""Finds rows in a table.

    >>> row = TableRow('tr','|')
    >>> mo = row.regexp.search(' | one | two |\n|one|two \n')
    >>> mo.group(1)
    '| one | two '
    >>> mo.group(0)
    ' | one | two |\n'
    
    """

    def __init__(self, tag, token):
        super(TableRow,self).__init__(tag,token)
        self.regexp = re.compile(self.re_string(),re.MULTILINE)

    def re_string(self):
        whitespace = r'[ \t]*'
        content = '(' + re.escape(self.token) + '.*?)'
        trailing_token = re.escape(self.token) + '?'
        return '^' + whitespace + content + trailing_token + \
               whitespace + r'(\n|\Z)'


class TableCell(WikiElement):

    r"""Finds cells in a table row.

    >>> cell = TableCell('td','|')
    >>> mo = cell.regexp.search('| one | two ')
    >>> mo.group(1)
    'one'
    >>> mo.group(0)
    '| one '
    
    """

    def __init__(self, tag, token):
        super(TableCell,self).__init__(tag,token )
        self.regexp = re.compile(self.re_string())

    def re_string(self):
        whitespace = r'[ \t]*'
        content = '(.*?)'
        look_ahead = '((?=' + esc_neg_look + re.escape(self.token[0]) + ')|$)'
        return esc_neg_look + re.escape(self.token) + whitespace + \
               content + whitespace + look_ahead    



##class Link(InlineElement):
##
##    """Finds and builds links."""
##    
##    def __init__(self,tag, token):
##        super(Link,self).__init__(tag,token)
##
##    def _build(self,mo,element_store, environ):
##        
##        for tag in self.child_elements:
##            m = tag.regexp.search(mo.group(1))
##            if m:
##                link = tag._build(m,element_store, environ)
##                if link:
##                    break
##        else:
##            link = None
##
##        if link:
##            return bldr.tag(link)
##        else:
##            return mo.group(0)

class Image(InlineElement):

    """Processes image elements.

    >>> img = Image('img',('{{','}}'), delimiter='|')
    >>> mo = img.regexp.search('{{ picture.jpg | An image of a house }}')
    >>> img._build(mo,{},None).generate().render()
    '<img src="picture.jpg" alt="An image of a house" title="An image of a house"/>'

    """

    def __init__(self, tag, token, delimiter):
        super(Image,self).__init__(tag,token )
        self.regexp = re.compile(self.re_string())
        self.delimiter = delimiter
        self.src_regexp = re.compile(r'^\s*(\S+)\s*$')

    def _build(self,mo,element_store, environ):
        body = mo.group(1).split(self.delimiter,1)
        src_mo = self.src_regexp.search(body[0])
        if not src_mo:
            return bldr.tag.span('Bad Image src')
        if sanitizer.is_safe_uri(src_mo.group(1)):
            link = src_mo.group(1)
        else:
            link = "unsafe_uri_detected"
        if len(body) == 1:
            alias = link
        else:
            alias = body[1].strip()
        return bldr.tag.__getattr__(self.tag)(src=link ,alt=alias, title=alias)


class NoWikiElement(InlineElement):

    """Inline no-wiki.

    When two or more end tokens are found together, only last marks
    the end of the element.
    
    """

    def __init__(self, tag, token):
        super(NoWikiElement,self).__init__(tag,token )
        self.regexp = re.compile(self.re_string(),re.DOTALL) 

    def _build(self,mo,element_store, environ):
        if self.tag:
            return bldr.tag.__getattr__(self.tag)(
                   fragmentize(mo.group(1), self.child_elements,
                               element_store,environ,
                               remove_escapes=False))
        else:
            return bldr.tag(fragmentize(mo.group(1),self.child_elements,
                                        element_store, environ,
                                        remove_escapes=False))

    def re_string(self):
        if isinstance(self.token,str):
            content = '(.+?' + re.escape(self.token[-1]) + '*)'
            return esc_neg_look + re.escape(self.token) + \
                   content + re.escape(self.token)
        else:
            content = '(.+?' + re.escape(self.token[1][-1]) + '*)'
            return esc_neg_look + re.escape(self.token[0]) + \
                   content + re.escape(self.token[1])


class PreBlock(BlockElement):
    """A preformatted block.

    If a closing token is found on a line with a space as the first
    character, the space will be removed from the output.
    
    """

    def __init__(self, tag, token ):
        super(PreBlock,self).__init__(tag,token )
        self.regexp = re.compile(self.re_string(),re.DOTALL+re.MULTILINE)
        self.regexp2 = re.compile(self.re_string2(),re.MULTILINE)

    def re_string(self):
        if isinstance(self.token,str):
            return '^' + re.escape(self.token) + r'\s*?\n(.*?\n)' + \
                   re.escape(self.token) + r'\s*?\n'
        else:
            start = '^' + re.escape(self.token[0]) + r'\s*?\n'
            content = r'(.+?\n)'
            end = re.escape(self.token[1]) + r'\s*?$'
            return start + content + end

    def re_string2(self):
        """Finds a closing token with a space at the start of the line."""
        if isinstance(self.token,str):
            return r'^ (\s*?' + re.escape(self.token) + r'\s*?\n)'
        else:
            return r'^ (\s*?' + re.escape(self.token[1]) + r'\s*?\n)'

    def _build(self,mo,element_store, environ):
        match = self.regexp2.sub(r'\1',mo.group(1))
        
        return bldr.tag.__getattr__(self.tag)(
            fragmentize(match,self.child_elements,
                        element_store, environ,remove_escapes=False))


class IndentedBlock(BlockElement):
    """An  indented block.

    """

    def __init__(self, tag, token):
        super(IndentedBlock,self).__init__(tag,token )
        self.regexp = re.compile(self.re_string(),re.MULTILINE)
        self.regexp2 = re.compile(self.re_string2(),re.MULTILINE)

    def re_string(self):
        return r'^((' + re.escape(self.token) \
               + r'.*?(\n|\Z))+)'

    def re_string2(self):
        """Finds a token at the start of the line."""
        return r'^' + re.escape(self.token)

    def _build(self,mo,element_store, environ):
        match = self.regexp2.sub(r'',mo.group(1)) # removes tokens during processing
        return bldr.tag.__getattr__(self.tag)(
            fragmentize(match,self.child_elements,
                        element_store, environ))


class LoneElement(BlockElement):
    """Element on a line by itself with no content (e.g., <hr/>)"""

    def __init__(self, tag, token):
        super(LoneElement,self).__init__(tag,token )
        self.regexp = re.compile(self.re_string(),re.DOTALL+re.MULTILINE)

    def re_string(self):
        return r'^(\s*?' + re.escape(self.token) + r'\s*?(\n|\Z))'

    def _build(self,mo,element_store, environ):
        return bldr.tag.__getattr__(self.tag)()

 
class BlankLine(WikiElement):

    """Blank lines divide elements but don't add any output."""

    def __init__(self):
        super(BlankLine,self).__init__(tag=None,token='' , child_elements=[])
        self.regexp = re.compile(self.re_string(),re.MULTILINE)

    def re_string(self):
        return r'^\s*(\Z|\n)' # r'^(\s*\n)+'
     
    def _build(self,mo,element_store, environ):
        return None

    def _process(self, mos, text, wiki_elements,element_store, environ):
        """Returns genshi Fragments (Elements and text)

        Custom _process method here just to avoid unnecessary calling of
        _build.
        """
        
        frags = []
        end = 0
        for mo in mos:
            if end != mo.start():
            # call again for leading text and extend the result list 
                frags.extend(fragmentize(text[end:mo.start()],wiki_elements[1:],
                                         element_store, environ))
            end = mo.end()
        # call again for trailing text and extend the result list
        if end < len(text):
            if not isinstance(wiki_elements[0],(list,tuple)):
                wiki_elements = wiki_elements[1:]
            frags.extend(fragmentize(text[end:],wiki_elements,
                 element_store, environ))

        return frags

class LineBreak(InlineElement):
    """An inline line break."""

    def __init__(self,tag, token, blog_style=False):
        self.blog_style = blog_style
        super(LineBreak,self).__init__(tag,token )
        self.regexp = re.compile(self.re_string(),re.DOTALL)

    def re_string(self):
        if self.blog_style:
            return '(' + esc_neg_look + re.escape(self.token) + r'\n?|\n(?!$))'
        else:
            return esc_neg_look + re.escape(self.token)
    
    def _build(self,mo,element_store, environ):
        return bldr.tag.__getattr__(self.tag)()
    
class GenericElement(InlineElement):
    """A generic element."""

    def __init__(self,pattern,tag,text_node=None,attrs=None):
        super(GenericElement,self).__init__(tag,pattern )
        self.attrs = attrs or {}
        self.text_node = text_node
        self.regexp = re.compile(self.re_string(),re.DOTALL)
        

    def re_string(self):
        escape = '(' + re.escape(escape_char) + ')*'
        return escape + self.token 
    
    def _build(self,mo,element_store, environ):
        d = mo.groupdict()
        if d.has_key('all') or d.has_key('environ'):
            raise Exception('GenericElement patterns cannot have "all" or \
                            "environ" as group names.')
        d['all'] = mo.group(0)
        d['environ'] = environ
        if mo.group(1):
            return mo.group(0)[1:]
        else:
            if isinstance(self.text_node,(Stream, bldr.Fragment,Markup)) or self.text_node is None:
                content = self.text_node
            elif isinstance(self.text_node, basestring):
                if sys.version >= '2.6':
                    content = self.text_node.format(**d)
                else:
                    content = P3Template(self.text_node).substitute(d)  
            else:
                content = self.text_node(mo, environ)

            if callable(self.attrs):
                attrs = self.attrs(mo, environ)
            else:
                attrs = self.attrs
            kwparams = {}
            for k,v in attrs.items():
                if isinstance(v,(Stream, bldr.Fragment,Markup)) or v is None:
                    kwparams[k+'_'] = v
                elif isinstance(v,basestring):
                    if sys.version >= '2.6':
                        kwparams[k+'_'] = v.format(**d)
                    else:
                        kwparams[k+'_'] = P3Template(v).substitute(d)
                else:
                    kwparams[k+'_'] =  v(mo, environ)
            if callable(self.tag):
                tag = self.tag(mo, environ)
            else:
                tag = self.tag
            if tag:
                return bldr.tag.__getattr__(tag)(content,**kwparams)
            else:
                return content
        
#############################################################################
# The WikeElement classes below are used for parsing macro argument strings # 
#############################################################################

class ArgString(WikiElement):
    """Base class for elements used on argument strings"""
    def __init__(self, tag='', token=''):
        super(ArgString,self).__init__(tag,token)
        self.regexp = re.compile(self.re_string(),re.DOTALL)


class KeywordArg(ArgString):
   """Finds keyword arguments"""

   def re_string(self):
      return r'(?P<key>\w[\w0-9]*) *'+re.escape(self.token) + \
               r' *(?P<body>.*?) *(?=\w[\w0-9]* *' + re.escape(self.token) +'|$)'

   def _build(self,mo,element_store, environ):
      if mo.group('body') == '':
         value = ''
      else:
         value = fragmentize(mo.group('body'),self.child_elements,
                             element_store, environ)
         if len(value) == 1:
             value = value[0]
         else:
             value = ImplicitList(value)
      name = mo.group('key')
      return (name, value)


class QuotedArg(InlineElement):
   """Finds quoted arguments"""

   def re_string(self):
      return esc_neg_look + r'(?P<quote>['+ re.escape(self.token) \
             +'])(?P<body>.*?)' + esc_neg_look + '(?P=quote)'

   def _build(self,mo,element_store, environ):
      if mo.group('body') == '':
          value = ''
      else:
          frags = fragmentize(mo.group('body'),self.child_elements,element_store, environ)
          assert len(frags) == 1
          value = frags[0]
      return value
         
class ListArg(ArgString):
   """Finds lists in argument strings 

   This is used for positional arguments only.
    
   """
    
   def re_string(self):
      return esc_neg_look + re.escape(self.token[0]) + r'(?P<body>.*?)' + esc_neg_look + re.escape(self.token[1])

   def _build(self,mo,element_store, environ):
      if mo.group('body') == '':
         value = []
      else:
         value = fragmentize(mo.group('body'),self.child_elements,element_store, environ)
      return value

class ExplicitListArg(ListArg):
    """Only finds lists where the string to be searched is fully enclosed

    This is used for keyword values in argument strings.
    
    """

    def re_string(self):
       return '^' + esc_neg_look + re.escape(self.token[0]) +r'(?P<body>.*?)' \
             + esc_neg_look+ re.escape(self.token[1]) + '$'


class WhiteSpace(ArgString):
   """Breaks up elements but doesn't add any output"""

   def re_string(self):
      return r'[ \n]+'

   def _build(self,mo,element_store, environ):
      return None



def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()    



# Module:   pagetypes
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Page Types and Handling

...
"""

import os
import csv
from StringIO import StringIO
from operator import itemgetter
from time import strftime, gmtime

try:
    import docutils.core
    HAS_DOCUTILS = True
except ImportError:
    HAS_DOCUTILS = False

from genshi.input import HTML
from genshi.builder import tag
from genshi.core import Markup
from genshi.filters import HTMLSanitizer
from genshi.output import HTMLSerializer

from mercurial.node import short

from circuits.web.exceptions import Redirect
from circuits.web.tools import expires, serve_file

from utils import NEWLINES
from highlight import highlight
from errors import NotImplementedErr, UnsupportedMediaTypeErr

class WikiPage(object):
    """Everything needed for rendering a page."""

    def __init__(self, environ, name, mime):
        super(WikiPage, self).__init__()

        self.environ = environ
        self.name = name
        self.mime = mime

        self.request = self.environ.request
        self.response = self.environ.response

        self.url = self.request.url
        self.render = self.environ.render

        self.search = self.environ.search
        self.storage = self.environ.storage

    def _get_text(self):
        return self.storage.page_text(self.name)

    def _get_page_data(self):
        text = self._get_text()
        rev, node, date, author, comment = self.storage.page_meta(self.name)

        base = os.path.basename(self.name)
        parent = os.path.dirname(self.name)

        data = {
            "rev": rev,
            "base": base,
            "date": date,
            "text": text,
            "parent": parent,
            "author": author,
            "name": self.name,
            "comment": comment,
            "node": short(node),
        }

        if hasattr(self.storage, "page_parent"):
            data["parent"] = self.storage.page_parent(self.name)

        self.environ.page = data

        return data

    def download(self):
        path = self.storage._file_path(self.name)
        expires(self.request, self.response, 60*60*24*30, force=True)
        return serve_file(self.request, self.response, path, type=self.mime)

    def edit(self):
        raise NotImplementedErr()

    def history(self):
        history = list(self.storage.page_history(self.name))[:30]
        data = {
            "name": self.name,
            "title": "History of %s" % self.name,
            "history": history,
            "strftime": strftime,
            "gmtime": gmtime,
            "ctxnav": list(self.environ._ctxnav("history"))
        }

        return self.render("history.html", **data)

    def view(self):
        raise NotImplementedErr()

class WikiPageText(WikiPage):
    """Pages of mime type text/* use this for display."""

    def _render(self, text=None):
        if text is None:
            text = self._get_text()

        return tag.pre(text)

    def edit(self):
        if not self.request.kwargs:
            if self.name in self.storage:
                data = {"page": self._get_page_data()}
                return self.render("edit.html", **data)
            else:
                data = {"page": {"name": self.name, "text": ""}}
                return self.render("edit.html", **data)

        action = self.request.kwargs.get("action", None)
        comment = self.request.kwargs.get("comment", "")
        parent = self.request.kwargs.get("parent", None)
        text = self.request.kwargs.get("text", "")

        text = NEWLINES.sub("\n", text)

        if action == "cancel":
            raise Redirect(self.url("/%s" % self.name))
        elif action == "preview":
            data = {
                "page": {"name": self.name, "text": text},
                "author": self.environ._user(),
                "comment": comment,
                "preview": True,
            }
            data["html"] = self._render(text)
            return self.render("edit.html", **data)
        elif action == "save":
            self.storage.reopen()
            self.search.update(self.environ)

            self.storage.save_text(self.name, text, self.environ._user(),
                comment, parent=parent)
            self.search.update_page(self, self.name, text=text)

            raise Redirect(self.url("/%s" % self.name))
        else:
            raise Exception("Invalid action %r" % action)

    def view(self):
        data = {
            "page": self._get_page_data(),
            "ctxnav": list(self.environ._ctxnav("view", self.name)),
        }
        data["html"] = self._render()
        return self.render("view.html", **data)

class WikiPageColorText(WikiPageText):
    """Text pages, but displayed colorized with pygments"""

    def view(self):
        data = {
            "page": self._get_page_data(),
            "ctxnav": list(self.environ._ctxnav("view", self.name)),
        }

        data["html"] = highlight(data["page"]["text"], mime=self.mime)
        return self.render("view.html", **data)

class WikiPageWiki(WikiPageText):
    """Pages of with wiki markup use this for display."""

    def _render(self, text=None, data={}):
        if text is None:
            text = self._get_text()

        return self.environ.parser.generate(text, context="block",
                environ=(self.environ, data))

    def edit(self):
        if not self.request.kwargs:
            if self.name in self.storage:
                data = {"page": self._get_page_data()}
                return self.render("edit.html", **data)
            else:
                data = {"page": {"name": self.name, "text": ""}}
                return self.render("edit.html", **data)

        action = self.request.kwargs.get("action", None)
        comment = self.request.kwargs.get("comment", "")
        parent = self.request.kwargs.get("parent", None)
        text = self.request.kwargs.get("text", "")

        text = NEWLINES.sub("\n", text)

        if action == "cancel":
            raise Redirect(self.url("/%s" % self.name))
        elif action == "preview":
            data = {
                "page": {"name": self.name, "text": text},
                "author": self.environ._user(),
                "comment": comment,
                "preview": True,
            }
            data["html"] = self._render(text, data)
            return self.render("edit.html", **data)
        elif action == "save":
            self.storage.reopen()
            self.search.update(self.environ)

            self.storage.save_text(self.name, text, self.environ._user(),
                    comment, parent=parent)
            self.search.update_page(self, self.name, text=text)

            raise Redirect(self.url("/%s" % self.name))
        else:
            raise Exception("Invalid action %r" % action)

    def view(self):
        data = {
            "page": self._get_page_data(),
            "ctxnav": self.environ._ctxnav("view", self.name),
        }
        data["html"] = self._render(data=data)
        return self.render("view.html", **data)

class WikiPageFile(WikiPage):
    """Pages of all other mime types use this for display."""

class WikiPageImage(WikiPageFile):
    """Pages of mime type image/* use this for display."""

    def view(self):
        if self.name not in self.storage:
            data = {"name": self.name}
            self.storage.reopen()
            self.search.update(self.environ)
            name, _ = os.path.splitext(self.name)
            data["results"] = sorted(self.search.find((name,)),
                    key=itemgetter(0), reverse=True)[:5]
            return self.render("notfound.html", **data)

        data = {
            "title": self.name,
            "name": self.name,
            "html": tag.img(
                alt=self.name,
                src=self.url("/+download/%s" % self.name)
            ),
            "ctxnav": list(self.environ._ctxnav("view", self.name)),
        }

        return self.render("view.html", **data)

class WikiPageCSV(WikiPageFile):
    """Display class for type text/csv."""

    def _render(self, text=None):
        if text is None:
            text = self._get_text()

        data = csv.reader(StringIO(text))

        rows = (tag.tr(tag.td(cell) for cell in row) for row in data)

        stream = tag.table(rows)

        return stream.generate()

    def edit(self):
        if not self.request.kwargs:
            if self.name in self.storage:
                data = {"page": self._get_page_data()}
                return self.render("edit.html", **data)
            else:
                data = {"page": {"name": self.name, "text": ""}}
                return self.render("edit.html", **data)

        action = self.request.kwargs.get("action", None)
        comment = self.request.kwargs.get("comment", "")
        parent = self.request.kwargs.get("parent", None)
        text = self.request.kwargs.get("text", "")

        text = NEWLINES.sub("\n", text)

        if action == "cancel":
            raise Redirect(self.url("/%s" % self.name))
        elif action == "preview":
            data = {
                "page": {"name": self.name, "text": text},
                "html": self._render(),
                "author": self.environ._user(),
                "comment": comment,
                "preview": True,
            }
            return self.render("edit.html", **data)
        elif action == "save":
            self.storage.reopen()
            self.search.update(self.environ)

            self.storage.save_text(self.name, text, self.environ._user(), comment,
                    parent=parent)
            self.search.update_page(self, self.name, text=text)

            raise Redirect(self.url("/%s" % self.name))
        else:
            raise Exception("Invalid action %r" % action)

    def view(self):
        data = {
            "page": self._get_page_data(),
            "html": self._render(),
            "ctxnav": list(self.environ._ctxnav("view", self.name)),
        }
        return self.render("view.html", **data)

class WikiPageRST(WikiPageText):
    """
    Display ReStructured Text.
    """

    def _render(self, text=None):
        if text is None:
            text = self._get_text()

        SAFE_DOCUTILS = dict(file_insertion_enabled=False, raw_enabled=False)

        return Markup(docutils.core.publish_parts(text, writer_name="html",
            settings_overrides=SAFE_DOCUTILS)["html_body"])

    def edit(self):
        if not self.request.kwargs:
            if self.name in self.storage:
                data = {"page": self._get_page_data()}
                return self.render("edit.html", **data)
            else:
                data = {"page": {"name": self.name, "text": ""}}
                return self.render("edit.html", **data)

        action = self.request.kwargs.get("action", None)
        comment = self.request.kwargs.get("comment", "")
        parent = self.request.kwargs.get("parent", None)
        text = self.request.kwargs.get("text", "")

        text = NEWLINES.sub("\n", text)

        if action == "cancel":
            raise Redirect(self.url("/%s" % self.name))
        elif action == "preview":
            data = {
                "page": {"name": self.name, "text": text},
                "html": self._render(text),
                "author": self.environ._user(),
                "comment": comment,
                "preview": True,
            }
            return self.render("edit.html", **data)
        elif action == "save":
            self.storage.reopen()
            self.search.update(self.environ)

            self.storage.save_text(self.name, text, self.environ._user(),
                    comment, parent=parent)
            self.search.update_page(self, self.name, text=text)

            raise Redirect(self.url("/%s" % self.name))
        else:
            raise Exception("Invalid action %r" % action)

    def view(self):
        if not HAS_DOCUTILS:
            raise UnsupportedMediaTypeErr("No docutils support available")

        data = {
            "page": self._get_page_data(),
            "html": self._render(),
            "ctxnav": list(self.environ._ctxnav("view", self.name)),
        }
        return self.render("view.html", **data)

class WikiPageHTML(WikiPageText):
    """Display HTML (genshi) templates"""

    def __init__(self, environ, name, mime):
        super(WikiPageHTML, self).__init__(environ, name, mime)

        self.sanitizer = HTMLSanitizer()
        self.serializer = HTMLSerializer()

    def _render(self, text=None):
        if text is None:
            text = self._get_text()

        return self.sanitizer(HTML(text))

    def edit(self):
        if not self.request.kwargs:
            if self.name in self.storage:
                data = {"page": self._get_page_data()}
                return self.render("edit.html", **data)
            else:
                data = {"page": {"name": self.name, "text": ""}}
                return self.render("edit.html", **data)

        action = self.request.kwargs.get("action", None)
        comment = self.request.kwargs.get("comment", "")
        parent = self.request.kwargs.get("parent", None)
        text = self.request.kwargs.get("text", "")

        text = NEWLINES.sub("\n", text)

        if action == "cancel":
            raise Redirect(self.url("/%s" % self.name))
        elif action == "preview":
            data = {
                "page": {"name": self.name, "text": text},
                "author": self.environ._user(),
                "comment": comment,
                "preview": True,
            }
            data["html"] = self._render(text)
            return self.render("edit.html", **data)
        elif action == "save":
            self.storage.reopen()
            self.search.update(self.environ)

            self.storage.save_text(self.name, text, self.environ._user(),
                comment, parent=parent)
            self.search.update_page(self, self.name, text=text)

            raise Redirect(self.url("/%s" % self.name))
        else:
            raise Exception("Invalid action %r" % action)

    def view(self):
        data = {
            "page": self._get_page_data(),
            "ctxnav": list(self.environ._ctxnav("view", self.name)),
        }
        data["html"] = self._render()
        return self.render("view.html", **data)



# Module:   include
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Include macro

Macro for inclusion of other wiki pages
"""

from genshi.builder import tag
from genshi.output import HTMLSerializer

serializer = HTMLSerializer()

def include(macro, environ, data, *args, **kwargs):
    """Include the contents of another wiki page.
    
    This macro allows you to include the contents of another wiki page,
    optinoally allowing you to include it parsed (//the default//) or
    unparsed (//parse=False//). If the specified page does not exist,
    then "Page Not Found" will be displayed.

    **Arguments:**
    * name=None (//the name of the page to include//)
    * parse=True (//whether to parse the page//)
    * source=False (//whether to display the genereated HTML / source//)

    **Example(s):**
    {{{
    <<include "SandBox">>
    }}}

    <<include "SandBox">>

    {{{
    <<include "SandBox", parse=False>>
    }}}

    <<include "SandBox", parse=False>>
    """
    
    name = kwargs.get("name", (args and args[0]) or None)

    if name is None:
        return None

    parse = kwargs.get("parse", True)
    source = kwargs.get("source", False)

    contents = environ.include(name, parse, data=data)

    if source:
        return tag.pre("".join(serializer(contents)))
    else:
        return contents



#!/usr/bin/env python

import os

try:
    from setuptools import setup, find_packages
    HAS_SETUPTOOLS = True
except ImportError:
    from distutils.core import setup
    HAS_SETUPTOOLS = False

if not HAS_SETUPTOOLS:
    import os
    from distutils.util import convert_path

    def find_packages(where=".", exclude=()):
        """Borrowed directly from setuptools"""

        out = []
        stack = [(convert_path(where), "")]
        while stack:
            where, prefix = stack.pop(0)
            for name in os.listdir(where):
                fn = os.path.join(where, name)
                if ("." not in name and os.path.isdir(fn) and 
                        os.path.isfile(os.path.join(fn, "__init__.py"))):
                    out.append(prefix+name)
                    stack.append((fn, prefix + name + "."))

        from fnmatch import fnmatchcase
        for pat in list(exclude) + ["ez_setup"]:
            out = [item for item in out if not fnmatchcase(item, pat)]

        return out

path = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(path, "README.rst")).read()
    RELEASE = open(os.path.join(path, "RELEASE.rst")).read()
except IOError:
    README = RELEASE = ""

setup(
    name="sahriswiki",
    description="A Lightweight Wiki / CMS / Blogging Engine using circuits.web",
    long_description="%s\n\n%s" % (README, RELEASE),
    author="James Mills",
    author_email="James Mills, prologic at shortcircuit dot net dot au",
    url="http://bitbucket.org/prologic/sahriswiki/",
    download_url="http://bitbucket.org/prologic/sahriswiki/downloads/",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Plugins",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 2.6",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application"],
    license="MIT",
    keywords="wiki cms blog engine circuits",
    platforms="POSIX",
    packages=find_packages("."),
    package_data={
        "sahriswiki": [
            "templates/*.html",
            "htdocs/css/*.css",
            "htdocs/img/*.png",
            "htdocs/favicon.ico",
            "htdocs/img/icons/*.png",]},
    include_package_data=True,
    scripts=["scripts/sahriswiki"],
    entry_points="""
    [console_scripts]
    sahriswiki = sahriswiki.main:main
    """,
    install_requires = [
        "circuits",
        "docutils",
        "genshi",
        "sqlalchemy",
        "pygments",
        "feedformatter",
    ],
    setup_requires=("hgtools",),
    use_hg_version={"increment": "0.01"},
)



# Module:   root
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Root Controller

...
"""

import os
import re
from urlparse import urlparse
from operator import itemgetter
from difflib import unified_diff
from time import gmtime, strftime

from genshi.core import Markup

from circuits.web.tools import check_auth, basic_auth
from circuits.web.controllers import expose, BaseController

import schema
from feedformatter import Feed
from highlight import highlight
from errors import ForbiddenErr, NotFoundErr

class Root(BaseController):

    def __init__(self, environ):
        super(Root, self).__init__()

        self.environ = environ

        self.config = self.environ.config
        self.render = self.environ.render
        self.search = self.environ.search
        self.storage = self.environ.storage

    @expose("index")
    def index(self, *args, **kwargs):
        if args:
            name = os.path.sep.join(args)
        else:
            for name in self.config.get("indexes"):
                if name in self.storage:
                    break
            if not name in self.storage:
                name = self.config.get("frontpage")

        page = self.environ.get_page(name)

        try:
            return page.view()
        except NotFoundErr:
            data = {"title": name}
            data["results"] = sorted(self.search.find((name,)),
                    key=itemgetter(0), reverse=True)[:5]
            if hasattr(self.storage, "page_parent"):
                data["parent"] = self.storage.page_parent(name)
            return self.render("notfound.html", **data)

    @expose("+download")
    def download(self, *args, **kwargs):
        name = os.path.sep.join(args) if args else self.config.get("index")
        page = self.environ.get_page(name)
        return page.download()

    @expose("+upload")
    def upload(self, *args, **kwargs):
        if not self.environ._login() and self.config.get("readonly"):
            raise ForbiddenErr("This wiki is in readonly mode.")

        action = kwargs.get("action", None)

        data = {
            "title": "Upload",
        }

        if action == "upload":
            file = kwargs.get("file", None)
            if file is not None:
                filename = file.filename
                filedata = file.value
                comment = kwargs.get("comment", "Uploaded file: %s" % filename)

                data["filename"] = filename

                author = self.cookie.get("username")
                if author:
                    author = author.value
                else:
                    author = self.request.headers.get("X-Forwarded-For",
                            self.request.remote.ip or "AnonymousUser")

                self.storage.reopen()
                self.storage.save_data(filename, filedata, author, comment)

        return self.render("upload.html", **data)

    @expose("+edit")
    def edit(self, *args, **kwargs):
        if not self.environ._login() and self.config.get("readonly"):
            raise ForbiddenErr("This wiki is in readonly mode.")

        name = os.path.sep.join(args)
        page = self.environ.get_page(name)
        return page.edit()

    @expose("+login")
    def login(self):
        db = self.environ.dbm.session
        users = dict(((user.username, user.password) \
                for user in db.query(schema.User).all()))
        realm = self.environ.config.get("name")

        if not check_auth(self.request, self.response, realm, users):
            return basic_auth(self.request, self.response, realm, users)

        if not self.request.login in users:
            return basic_auth(self.request, self.response, realm, users)

        self.request.session["login"] = self.request.login

        referer = self.request.headers.get("Referer", None)
        if referer:
            base = urlparse(self.url())
            link = urlparse(referer)
            if all([base[i] == link[i] for i in range(2)]):
                return self.redirect(referer)

        data = {
            "title": "Login",
            "html": Markup("Login successful.")
        }

        return self.render("view.html", **data)

    @expose("+logout")
    def logout(self):
        db = self.environ.dbm.session
        users = dict(((user.username, user.password) \
                for user in db.query(schema.User).all()))
        realm = self.environ.config.get("name")

        if "login" in self.request.session:
            del self.request.session["login"]

        if "Authorization" in self.request.headers:
            del self.request.headers["Authorization"]

        return basic_auth(self.request, self.response, realm, users)

    @expose("+about")
    def about(self):
        return self.render("about.html")

    @expose("+search")
    def search(self, *args, **kwargs):

        def snippet(name, words):
            """Extract a snippet of text for search results."""

            try:
                text = unicode(self.storage.open_page(name).read(),
                    "utf-8", errors="replace")
            except NotFoundErr:
                return u""

            regexp = re.compile(u"|".join(re.escape(w) for w in words),
                    re.U | re.I)
            match = regexp.search(text)
            if match is None:
                return u""
            position = match.start()
            min_pos = max(position - 60, 0)
            max_pos = min(position + 60, len(text))
            highlighted = "<span class=\"highlight\">%s</span>" % match.group(0)
            return regexp.sub(highlighted, text[min_pos:max_pos])

        def search(words):
            self.storage.reopen()
            self.search.update(self.environ)
            results = list(self.search.find(words))
            results.sort(key=itemgetter(0), reverse=True)
            for score, name in results:
                yield score, name, Markup(snippet(name, words))

        query = kwargs.get("q", None)
        if query is not None:
            query = query.strip()

        if not query:
            data = {
                "title": "Page Index",
                "ctxnav": self.environ._ctxnav("index"),
            }
            if hasattr(self.storage, "all_pages_tree"):
                data["pages"] = sorted(self.storage.all_pages_tree())
            else:
                data["pages"] = sorted(self.storage.all_pages())

            return self.render("index.html", **data)

        if query in self.storage:
            return self.redirect(query)

        words = tuple(self.search.split_text(query, stop=False))
        if not words:
            words = (query,)

        data = {
            "title": "Search results for \"%s\"" % query,
            "name": "Search",
            "query": query,
            "results": list(search(words)),
            "ctxnav": self.environ._ctxnav("search"),
        }

        return self.render("search.html", **data)

    @expose("+backlinks")
    def backlinks(self, *args, **kwargs):
        name = os.path.sep.join(args)

        self.storage.reopen()
        self.search.update(self.environ)

        data = {
            "title": "BackLinks for \"%s\"" % name,
            "pages": sorted(self.search.page_backlinks(name),
                key=itemgetter(0)),
            "ctxnav": self.environ._ctxnav("search"),
        }
        return self.render("index.html", **data)

    @expose("+feed")
    def feed(self, *args, **kwargs):
        name = os.path.sep.join(args) if args else None
        format = kwargs.get("format", "rss1")

        if not format in ("rss1", "rss2", "atom"):
            raise Exception("Invalid format %r" % format)

        feed = Feed()

        if name is not None:
            feed.feed["title"] = "%s :: %s" % (name, self.environ.site["name"])
        else:
            feed.feed["title"] = self.environ.site["name"]

        feed.feed["link"] = self.request.server.base
        feed.feed["author"] = self.environ.site["author"]
        feed.feed["description"] = self.environ.site["description"]

        if name is not None:
            history = list(self.storage.page_history(name))[:30]
            for rev, date, author, comment in history:
                item = {}
                item["title"] = "%s by %s" % (name, author)
                item["link"] = self.request.url(name)
                item["description"] = comment
                item["pubDate"] = date
                item["guid"] = str(rev)

                feed.items.append(item)
        else:
            history = list(self.storage.history())[:30]
            for name, rev, date, author, comment in history:
                item = {}
                item["title"] = "%s by %s" % (name, author)
                item["link"] = self.request.url(name)
                item["description"] = comment
                item["pubDate"] = date
                item["guid"] = str(rev)

                feed.items.append(item)

        self.response.headers["Content-Type"] = "application/xml"
        return getattr(feed, "format_%s_string" % format)()

    @expose("+orphaned")
    def orphaned(self, *args, **kwargs):
        self.storage.reopen()
        self.search.update(self.environ)

        data = {
            "title": "Orphaned Pages",
            "pages": sorted(self.search.orphaned_pages(), key=itemgetter(0)),
            "ctxnav": self.environ._ctxnav("index"),
        }

        return self.render("orphaned.html", **data)

    @expose("+wanted")
    def wanted(self, *args, **kwargs):
        self.storage.reopen()
        self.search.update(self.environ)

        data = {
            "title": "Wanted Pages",
            "pages": sorted(self.search.wanted_pages(),
                key=itemgetter(0), reverse=True),
            "ctxnav": self.environ._ctxnav("index"),
        }

        return self.render("wanted.html", **data)

    @expose("+history")
    def history(self, *args, **kwargs):
        name = os.path.sep.join(args) or None
        if name:
            page = self.environ.get_page(name)
            return page.history()

        history = list(self.storage.history())[:30]

        data = {
            "title": "Recent Changes",
            "history": history,
            "strftime": strftime,
            "gmtime": gmtime,
            "ctxnav": self.environ._ctxnav("history"),
        }

        return self.render("recentchanges.html", **data)

    @expose("facicon.ico")
    def favicon(self, *args, **kwargs):
        if "favicon.ico" in self.storage:
            return self.download("favicon.ico")
        else:
            return self.serve_file(
                os.path.join(self.config.get("theme"), "htdocs", "favicon.ico")
            )

    @expose("robots.txt")
    def robots(self, *args, **kwargs):
        if "robots.txt" in self.storage:
            return self.download("robots.txt")

        self.response.headers["Content-Type"] = "text/plain"

        return "\r\n".join((
            "User-agent: *",
            "Disallow: /+",
            "Disallow: /%2b",
            "Disallow: /%2B",
        ))

    @expose("+delete")
    def delete(self, *args, **kwargs):
        name = os.path.sep.join(args)

        if name not in self.storage:
            data = {"title": name}
            data["results"] = sorted(self.search.find((name,)),
                    key=itemgetter(0), reverse=True)[:5]
            if hasattr(self.storage, "page_parent"):
                data["parent"] = self.storage.page_parent(name)
            return self.render("notfound.html", **data)

        action = kwargs.get("action", None)

        if not action:
            data = {"title": name}
            return self.render("delete.html", **data)

        if action == "delete":
            comment = kwargs.get("comment", "")

            self.storage.reopen()
            self.search.update(self.environ)

            self.storage.delete_page(name, self.environ._user(), comment)
            self.search.update_page(self, name)

            data = {"success": True, "message": "Page deleted successfully."}
            return self.render("delete.html", **data)
        else:
            raise Exception("Invalid action %r" % action)

    @expose("+rename")
    def rename(self, *args, **kwargs):
        name = os.path.sep.join(args)

        if name not in self.storage:
            data = {"title": name}
            data["results"] = sorted(self.search.find((name,)),
                    key=itemgetter(0), reverse=True)[:5]
            if hasattr(self.storage, "page_parent"):
                data["parent"] = self.storage.page_parent(name)
            return self.render("notfound.html", **data)

        action = kwargs.get("action", None)

        if not action:
            data = {"title": name}
            return self.render("rename.html", **data)

        if action == "rename":
            newname = kwargs.get("name", "")
            if newname and newname not in self.storage:
                comment = kwargs.get("comment", "")

                self.storage.reopen()
                self.search.update(self.environ)

                user = self.environ._user()

                text = self.storage.page_text(name)
                self.storage.save_text(newname, text, user, comment)
                self.search.update_page(self, newname, text=text)

                self.storage.delete_page(name, user, comment)
                self.search.update_page(self, name)

                data = {
                    "success": True,
                    "message": "Page renamed successfully.",
                }
            elif newname in self.storage:
                data = {
                    "success": False,
                    "message": "%s already exists" % newname,
                }
            else:
                data = {
                    "success": False,
                    "message": "A new name is requried",
                }

            return self.render("rename.html", **data)
        else:
            raise Exception("Invalid action %r" % action)

    @expose("+diff")
    def diff(self, *args, **kwargs):
        name = os.path.sep.join(args)

        rev = kwargs["rev"]

        to_rev = int(rev)
        from_rev = to_rev - 1

        text = self.storage.revision_text(name, from_rev).split("\n")
        other = self.storage.revision_text(name, to_rev).split("\n")

        to_date = ""
        from_date = ""
        date_format = "%Y-%m-%dT%H:%M:%SZ"

        for history in self.storage.page_history(name):
            if history[0] == to_rev:
                to_date = history[1]
            elif history[0] == from_rev:
                from_date = history[1]

        if not from_date:
            from_date = to_date

        diff = "\n".join(unified_diff(text, other,
            "%s@%d" % (name, from_rev),
            "%s@%d" % (name, to_rev),
            strftime(date_format, gmtime(from_date)),
            strftime(date_format, gmtime(to_date))))

        data = {
            "title": "diff of %s from %s to %s" % (name, from_rev, to_rev),
            "diff": highlight(diff, lang="diff"),
        }

        return self.render("diff.html", **data)



# Module:   highlight
# Date:     17th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Highlighting

...
"""

import pygments
import pygments.util
import pygments.lexers
import pygments.formatters

from genshi import Markup
from genshi.builder import tag

class HTMLFormatter(pygments.formatters.HtmlFormatter):

    def wrap(self, source, outfile):
        return self._wrap_code(source)

    def _wrap_code(self, source):
        yield 0, "<pre xml:space=\"preserve\">"
        for i, t in source:
            if not t.strip():
                t = "<br />"
            yield i, t
        yield 0, "</pre>"

def highlight(text, mime=None, lang=None, linenos=False):
    formatter = HTMLFormatter(cssclass="code", linenos=linenos)

    try:
        if mime:
            lexer = pygments.lexers.get_lexer_for_mimetype(mime)
        elif lang:
            lexer = pygments.lexers.get_lexer_by_name(lang)
        else:
            lexer = pygments.lexers.guess_lexer(text)
    except pygments.util.ClassNotFound:
        return tag.pre(text)

    return Markup(pygments.highlight(text, lexer, formatter))



# Module:   tools
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Tools and Filters

...
"""

import os
import signal
from hashlib import md5
from marshal import dumps

from genshi import Markup

from mercurial.node import short

from circuits.web import Response
from circuits.web.tools import gzip
from circuits import handler, BaseComponent
from circuits.web.tools import validate_etags

import sahriswiki

class CacheControl(BaseComponent):

    channel = "web"

    def __init__(self, environ):
        super(CacheControl, self).__init__()

        self.environ = environ

    @handler("request", filter=True, priority=1.0)
    def _on_request(self, request, response):
        if request.path in ("/+login", "/+logout"):
            return

        repo = short(self.environ.storage.repo_node())
        sess = md5(dumps(request.session)).hexdigest()
        config = md5(dumps(self.environ.config.copy())).hexdigest()
        version = sahriswiki.__version__

        etag = "%s/%s/%s/%s" % (repo, sess, config, version)

        response.headers.add_header("ETag", etag)
        response = validate_etags(request, response)
        if response:
            return response

class Compression(BaseComponent):

    channel = "web"

    mime_types = ["text/plain", "text/html", "text/css"]

    @handler("response_started")
    def _on_response_started(self, event, response_event):
        response = response_event[0]
        gzip(response, mime_types=self.mime_types)

class ErrorHandler(BaseComponent):

    channel = "web"

    def __init__(self, environ):
        super(ErrorHandler, self).__init__()

        self.environ = environ
        self.render = self.environ.render

    @handler("httperror", filter=True)
    def _on_httperror(self, event, request, response, code, **kwargs):
        self.environ.request = request
        self.environ.response = response
        data = event.data.copy()
        data["title"] = "Error"
        data["traceback"] = Markup(data["traceback"])
        data["description"] = Markup(data["description"] or u"")
        response.body = self.render("error.html", **data)
        return self.push(Response(response))

class SignalHandler(BaseComponent):

    def __init__(self, environ):
        super(SignalHandler, self).__init__()

        self.environ = environ

    @handler("signal", target="*")
    def _on_signal(self, sig, stack):
        if os.name == "posix" and sig == signal.SIGHUP:
            self.storage.reopen()
            self.environ.config.reload_config()



# Module:   env
# Date:     12th July 2010
# Author:   James Mills, prologic at shortcircuit dot net dot au

"""Environment Container

...
"""

import os
from hashlib import md5
from urllib import basejoin
from itertools import chain
from urlparse import urlparse

from circuits import handler, BaseComponent

from genshi.builder import tag
from genshi.template import TemplateLoader

from creoleparser import create_dialect, creole11_base, Parser

import macros
import sahriswiki
from utils import page_mime
from auth import Permissions
from search import WikiSearch
from dbm import DatabaseManager
from storage import WikiSubdirectoryIndexesStorage as DefaultStorage

from pagetypes import WikiPageText, WikiPageHTML
from pagetypes import WikiPageWiki, WikiPageFile, WikiPageImage
from pagetypes import WikiPageColorText, WikiPageCSV, WikiPageRST

class Environment(BaseComponent):

    filename_map = {
        "README":       (WikiPageText,  "text/plain"),
        "COPYING":      (WikiPageText,  "text/plain"),
        "CHANGES":      (WikiPageText,  "text/plain"),
        "MANIFEST":     (WikiPageText,  "text/plain"),
        "LICENSE":      (WikiPageText,  "text/plain"),
        "ChangeLog":    (WikiPageText,  "text/plain"),
        "favicon.ico":  (WikiPageImage, "image/x-icon"),
    }

    mime_map = {
        "text":                     WikiPageText,
        "application/javascript":   WikiPageColorText,
        "text/x-python":            WikiPageColorText,
        "text/css":                 WikiPageColorText,
        "text/csv":                 WikiPageCSV,
        "text/html":                WikiPageHTML,
        "text/x-rst":               WikiPageRST,
        "text/x-wiki":              WikiPageWiki,
        "image":                    WikiPageImage,
        "":                         WikiPageFile,
    }

    def __init__(self, config):
        super(Environment, self).__init__()

        self.config = config

        self.dbm = DatabaseManager(self.config.get("db"),
            echo=(self.config.get("debug") and self.config.get("verbose")),
        ).register(self)

        self.storage = DefaultStorage(
            self.config.get("repo"),
            self.config.get("encoding"),
            index=self.config.get("index"),
            indexes=self.config.get("indexes"),
        )

        self.search = WikiSearch(
            self.dbm.session,
            self.config.get("language"),
            self.storage,
        )

        self.parser = Parser(
            create_dialect(
                creole11_base,
                macro_func=macros.dispatcher,
                wiki_links_base_url="/",
                wiki_links_class_func=self._wiki_links_class_func,
                wiki_links_path_func=self._wiki_links_path_func,
            ),
            method="xhtml"
        )

        template_config = {
            "allow_exec": False,
            "auto_reload": True,
            "default_encoding": self.config.get("encoding"),
            "search_path": [
                self.storage.path,
                os.path.join(self.config.get("theme"), "templates"),
            ],
            "variable_lookup": "lenient",
        }
            
        self.templates = TemplateLoader(**template_config)

        self.macros = macros.loadMacros()

        self.stylesheets = []
        self.version = sahriswiki.__version__

        self.site = {
            "name": self.config.get("name"),
            "author": self.config.get("author"),
            "keywords": self.config.get("keywords"),
            "description": self.config.get("description"),
        }

        self.request = None
        self.response = None

    def _login(self):
        return self.request.session.get("login", self.request.login)

    def _user(self):
        return self._login() or self.request.headers.get(
                "X-Forwarded-For", self.request.remote.ip)

    def _permissions(self):
        return Permissions(self, self._login())

    def _metanav(self):
        yield ("About",       self.url("/+about"),    )
        yield ("Help",        self.url("/Help"),      )
        yield ("History",     self.url("/+history"),  )

        if not self._login():
            yield ("Login",   self.url("/+login"),    )
        else:
            yield ("Logout",  self.url("/+logout"),   )
            yield ("Profile", self.url("/+profile"),  )

        yield ("Register",    self.url("/+register"), )

    def _ctxnav(self, type="view", name=None):
        permissions = self._permissions()
        if name and type == "view":
            yield ("Functions",     self._ctxnav("func", name),)
            yield ("Information",   self._ctxnav("info", name),)
            yield ("Miscellaneous", self._ctxnav("misc", name),)
        elif type in ("index", "search"):
            yield ("Index",         self.url("/+search"))
            yield ("Orphaned",      self.url("/+orphaned"))
            yield ("Wanted",        self.url("/+wanted"))
        elif type == "history":
            if name:
                yield ("RSS 1.0",   self.url("/+feed/%s/?format=rss1") % name)
                yield ("RSS 2.0",   self.url("/+feed/%s/?format=rss2") % name)
                yield ("Atom",      self.url("/+feed/%s/?format=atom") % name)
            else:
                yield ("RSS 1.0",   self.url("/+feed/?format=rss1"))
                yield ("RSS 2.0",   self.url("/+feed/?format=rss2"))
                yield ("Atom",      self.url("/+feed/?format=atom"))
        elif name and type == "func":
            if "PAGE_EDIT" in permissions:
                yield ("Edit",      self.url("/+edit/%s" % name))
            if "PAGE_DELETE" in permissions:
                yield ("Delete",    self.url("/+delete/%s" % name))
            if "PAGE_RENAME" in permissions:
                yield ("Rename",    self.url("/+rename/%s" % name))
        elif name and type == "info":
            yield ("History",       self.url("/+history/%s" % name))
            yield ("Feeds",          self._ctxnav("history", name))
        elif name and type == "misc":
            yield ("Download",      self.url("/+download/%s" % name))
            if "PAGE_UPLOAD" in permissions:
                yield ("Upload",    self.url("/+upload/%s" % name))

    def _breadcrumbs(self, page=None):
        yield ("", "Home", "Home",)
        if page and "name" in page:
            xs = []
            name = page["name"]
            if not name == self.config.get("frontpage"):
                parts = name.split("/")
                for x in parts[:-1]:
                    xs.append(x)
                    yield ("/".join(xs), x, x,)
                base = os.path.basename(name)
                yield ("+backlinks/%s" % name, base, "View BackLinks",)

    def _wiki_links_class_func(self, type, url, body, name):
        if type == "wiki":
            if url and url[0] == ".":
                url = url[1:]
            if url and url[0] in "./":
                url = url[1:]
            if url and url[0] == "/":
                url = url[1:]
            if url in self.storage:
                return "wiki"
            else:
                return "wiki new"
        elif type == "url":
            base = urlparse(self.url("/"))
            link = urlparse(url)
            if not all([base[i] == link[i] for i in range(2)]):
                return "external"

    def _wiki_links_path_func(self, tag, path, (environ, context)):
        if tag == "img":
            return self.url("/+download", path)
        elif tag == "a":
            if path.startswith(".."):
                path = path[2:]
                if path and path[0] == "/":
                    path = path[1:]
                dirname = os.path.dirname(context["page"]["name"])
                return os.path.join(dirname, path)
            elif path.startswith("."):
                path = path[1:]
                if path and path[0] == "/":
                    path = path[1:]
                return os.path.join(context["page"]["name"], path)
        return path

    def url(self, *args):
        return self.request.url("/".join(args))

    def staticurl(self, url):
        base = self.config.get("static-baseurl", None)
        if base:
            return basejoin(base, url)
        else:
            return self.request.url("/%s" % url)

    def get_page(self, name):
        """Creates a page object based on page"s mime type"""

        try:
            page_class, mime = self.filename_map[name]
        except KeyError:
            mime = page_mime(name)
            major, minor = mime.split("/", 1)
            try:
                page_class = self.mime_map[mime]
            except KeyError:
                try:
                    page_class = self.mime_map[major]
                except KeyError:
                    page_class = self.mime_map[""]

        return page_class(self, name, mime)

    def include(self, name, parse=True, context="block", data=None):
        if name in self.storage:
            text = self.storage.page_text(name)
            if parse:
                return self.parser.generate(text, context=context,
                        environ=(self, data))
            else:
                return tag.pre(text)
        else:
            return tag.div(tag.p(u"Page %s Not Found" % name), class_="error")

    def render(self, template, **data):
        data.update({
            "sahriswiki": {
                "version": sahriswiki.__version__
            },
            "url":         self.url,
            "site":        self.site,
            "include":     self.include,
            "config":      self.config,
            "staticurl":   self.staticurl,
            "permissions": self._permissions(),
            "ctxnav":      chain(self._ctxnav(), data.get("ctxnav", [])),
            "metanav":     chain(self._metanav(), data.get("metanav", [])),
            "breadcrumbs": list(self._breadcrumbs(data.get("page", None))),
        })
        t = self.templates.load(template)
        return t.generate(**data).render("xhtml", doctype="xhtml")

    @handler("request", priority=1.0, target="web")
    def _on_request(self, request, response):
        self.request = request
        self.response = response
